"""companda compares pandas DataFrames"""

import pandas as pd
from pandas.api.types import (
    is_integer_dtype,
    is_float_dtype,
    is_datetime64_any_dtype,
)
import numpy as np

from .helpers import get_column_dtype


class CompandaNotEqualError(Exception):
    def __init__(self, message):
        super().__init__(message)


class Companda(object):
    """class Companda is the output from companda(df1, df2); evaluates as boolean + stores a message"""

    def __init__(self,
                 equal: bool,
                 columns_equal: bool,
                 message='', ):
        self.equal = equal
        self.columns_equal = columns_equal
        self.message = message

    def __bool__(self):
        return self.equal

    def __repr__(self):
        return f'COMPANDA({self.equal}); columns_equal={self.columns_equal}; \n' \
               f'{self.message}'


def companda(df1: pd.DataFrame,
             df2: pd.DataFrame,
             epsilon=.001,
             check_dtype=False,
             ignore_all_nan_columns=False,
             ignore_index=False,
             ):
    """compare two DataFrames; return a Companda object that is truth-y or false-y.

    Args:
        df1: pd.DataFrame
        df2: pd.DataFrame
        epsilon: float (allowable decimal error)
        check_dtype: bool
        ignore_all_nan_columns: ignore any all NaN (i.e. empty) columns
        ignore_index: ignore values of index

    Returns: Companda(True) if and only if:
        1. columns are equal (i.e. both subsets of each other)
        2. indices are equal
        3. data is present/absent in same locations
        4. data is equal (within decimal error gamma)
    else: returns Companda(False)
    """
    if ignore_all_nan_columns:
        df1 = df1.copy()
        df2 = df2.copy()
        for df in [df1, df2]:
            for col in df.columns:
                if get_column_dtype(df[col], 'pd') is None:
                    df.drop([col], axis=1, inplace=True)
    # COLUMNS
    cols1 = set(df1.columns)
    cols2 = set(df2.columns)
    missing_from_2 = []
    for col in cols1:
        if col not in cols2:
            missing_from_2.append(col)
    missing_from_1 = []
    for col in cols2:
        if col not in cols1:
            missing_from_1.append(col)
    if missing_from_2 or missing_from_1:
        msg = f'Columns {missing_from_2} missing from df2 and {missing_from_1} missing from df1'
        return Companda(False, False, msg)

    if len(df1.columns) != len(df2.columns):
        return Companda(False, False, f'len(df1.cols) = {len(df1.columns)}, len(df2.cols) = {len(df2.columns)}')

    # INDEX
    df1 = df1.sort_index()
    df2 = df2.sort_index()

    if len(df1) != len(df2):
        return Companda(False, True, f'len(df1) = {len(df1)}, len(df2) = {len(df2)}')

    if not ignore_index:
        if df1.index.name != df2.index.name:
            return Companda(False, True,
                            f'Different index names: {df1.index.name}, {df2.index.name}')

        # coerce index dtype if we're not checking this
        # if not check_dtype:
        #     df1.index = df1.index.astype(type(df2.index))

        index_unequal = pd.Series(df1.index != df2.index)
        if index_unequal.sum():
            return Companda(False, True,
                            f'Equal length indices, but {index_unequal.sum()} out of {len(index_unequal)} '
                            f'index values are different. \n'
                            f'df1 index: {df1.index} \n'
                            f'df2 index: {df2.index}')

    # VALUES
    for col in df1.columns:
        # datatype checks
        if check_dtype and not df1[col].dtype is df2[col].dtype:
            return Companda(False, True,
                            f"columns and indices equal, but datatypes not equal in column {col}. \n"
                            f"types: {df1[col].dtype} / {df2[col].dtype}.")

        # CHECK NUMERIC DATATYPES
        if is_float_dtype(df1[col]) or is_integer_dtype(df1[col]):
            if np.array_equal(df1[col].isna(), df2[col].isna()):
                if np.array_equal(df1[col].dropna(), df2[col].dropna()):
                    continue
                else:
                    diff = np.abs(np.subtract(df1[col].dropna().values, df2[col].dropna().values))
                    columns_are_different = diff > epsilon
                if columns_are_different.sum() == 0:
                    continue
                else:
                    return Companda(False, True,
                                    f"Columns and indices equal; values not almost equal in column: {col}.\n"
                                    f"Maximum delta (absolute): {diff.max()}")
            else:
                print(df1)
                print(df2)
                return Companda(False, True,
                                f"Columns and indices equal; values have different NaN values in {col}.")
        elif is_datetime64_any_dtype(df1[col]):
            if df1[col].dt.tz != df2[col].dt.tz:
                return Companda(False, True,
                                f"unequal timezones: {df1[col].dt.tz}, {df2[col].dt.tz}")
            if np.array_equal(df1[col].isna(),
                              df2[col].isna()) and np.array_equal(df1[col].dropna().values,
                                                                  df2[col].dropna().values):
                continue
            else:
                print(df1)
                print(df2)
                return Companda(False, True,
                                f"columns and indices equal; datetime values different in {col}.")
        else:
            if np.array_equal(df1[col].isna(),
                              df2[col].isna()) and np.array_equal(df1[col].dropna(),
                                                                  df2[col].dropna()):
                continue
            else:
                for i in range(len(df1)):
                    if df1[col].iloc[i] != df2[col].iloc[i]:
                        print(i, df1[col].iloc[i], df2[col].iloc[i])
                return Companda(False, True,
                                f"columns and indices equal; values not equal in column {col}.\n"
                                f"df1[{col}]: {df1[col]} \n"
                                f"df2[{col}]: {df2[col]}")

    return Companda(True, True, f'EQUAL, checked_dtype={check_dtype}, ignore_index={ignore_index}')



from sqlalchemy import Table
from pandabase.helpers import *


def drop_db_table(table_name, con, schema=None):
    """Drop table [table_name] from con (or con.schema, if schema kwarg is supplied)
    utility function to avoid calling SQL/SQLA directly during maintenance, etc."""
    engine = engine_builder(con)
    meta = sqa.MetaData(engine)
    meta.reflect(bind=engine, schema=schema)

    if schema is None:
        table_namespace = table_name
    else:
        table_namespace = f'{schema}.{table_name}'

    t = meta.tables[table_namespace]

    with engine.begin():
        t.drop()


def get_db_table_names(con, schema=None):
    """get a list of table names from con (or con.schema, if schema kwarg is supplied)"""
    meta = sqa.MetaData()
    engine = engine_builder(con)
    meta.reflect(bind=engine, schema=schema)
    return list(meta.tables.keys())


def get_table_column_names(con, table_name, schema=None):
    """get a list of column names from con, table_name (or con.schema if schema kwarg is supplied)"""
    meta = sqa.MetaData()
    engine = engine_builder(con)
    meta.reflect(bind=engine, schema=schema)

    if schema is None:
        table_namespace = table_name
    else:
        table_namespace = f'{schema}.{table_name}'

    return list(meta.tables[table_namespace].columns)


def describe_database(con, schema=None):
    """
    Returns a description of con (or con.schema if schema kwarg is supplied): {table_names: {table_info_dicts}}

    Args:
        con: string URI or db engine
        schema: Specify the schema (if database flavor supports this). If None, use default schema.

    Returns:
        {'table_name_1': {'min': min, 'max': max, 'count': count},
         ... }
    """
    engine = engine_builder(con)
    meta = sqa.MetaData()
    meta.reflect(bind=engine, schema=schema)

    res = {}

    for table_name in meta.tables:
        try:
            table = Table(table_name, meta, autoload=True, autoload_with=engine)
            index = table.primary_key.columns
            if len(index) == 1:
                index = index.items()[0][0]
                minim = engine.execute(sqa.select([sqa.func.min(sqa.text(index))]).select_from(table)).scalar()
                maxim = engine.execute(sqa.select([sqa.func.max(sqa.text(index))]).select_from(table)).scalar()
                count = engine.execute(sqa.select([sqa.func.count()]).select_from(table)).scalar()
                res[table_name] = {'min': minim, 'max': maxim, 'count': count}
            else:
                count = engine.execute(sqa.select([sqa.func.count()]).select_from(table)).scalar()
                res[table_name] = {'index_type': 'multi', 'index_cols': str(index.keys()), 'count': count}

        except Exception as e:
            print(f'failed to describe table: {table_name} due to {e}')

    return res




import pandas as pd
import numpy as np
from pandas.api.types import (is_bool_dtype,
                              is_datetime64_any_dtype,
                              is_integer_dtype,
                              is_float_dtype,
                              is_object_dtype,
                              )

import sqlalchemy as sqa
from sqlalchemy import Column, Integer, String, Float, DateTime, Boolean, TIMESTAMP

# fake random index name in case of not explicitly indexed data
PANDABASE_DEFAULT_INDEX = 'pandabase_auto_generated_index'


def _sqa_type2pandas_type(sqa_dtype, index=False):
    """explicitly map SQLAlchemy datatypes to our Pandas dtypes. param index is to prevent nullable index"""
    if sqa_dtype == Integer:
        if index:
            return int
        else:
            return pd.Int64Dtype()
    elif sqa_dtype == Float:
        return np.float64
    elif sqa_dtype == Boolean:
        return np.bool_
    elif sqa_dtype == String:
        return np.str_
    elif sqa_dtype in [DateTime, TIMESTAMP] or isinstance(sqa_dtype, TIMESTAMP):
        return np.datetime64
    else:
        raise TypeError(f'Unknown sqlalchemy dtype: {sqa_dtype}')


def series_is_boolean(col: pd.Series or pd.Index):
    """
    returns:
        None if column is all None;
        True if a pd.Series only contains True, False, and None;
        False otherwise

    caveat: does not interpret all-zero or all-one columns as boolean"""
    if len(col.unique()) == 1 and col.unique()[0] is None:
        # return None for all-None columns
        return None
    elif col.isna().all():
        return None
    elif is_bool_dtype(col):
        return True
    elif is_object_dtype(col):
        for val in col.unique():
            if val not in [True, False, None]:
                return False
        if not (False in col.unique() and True in col.unique()):
            return False
        return True
    elif is_integer_dtype(col) or is_float_dtype(col):
        for val in col.unique():
            if pd.isna(val):
                continue
            if val not in [1, 0, None]:
                return False
            if not (0 in col.unique() and 1 in col.unique()):
                return False
        return True
    return False


def engine_builder(con):
    """
    Returns a SQLAlchemy engine from a URI (if con is a string)
    if con is already a connection, return con without modifying it.
    """
    if isinstance(con, str):
        if con[:8].lower() == 'postgres':
            con = sqa.create_engine(con, connect_args={'connect_timeout': 10})
        else:
            con = sqa.create_engine(con)

    return con


def _get_type_from_df_col(col: pd.Series, index: bool):
    """
    Take a pd.Series, return its SQLAlchemy datatype
    If it doesn't match anything, return String
    Args:
        col: pd.Series to check
        index: if True, index cannot be boolean
    Returns:
        sqlalchemy Type or None
        one of {Integer, Float, Boolean, DateTime, String, or None (for all NaN)}
    """
    if col.isna().all():
        return None

    if is_bool_dtype(col):
        if index:
            raise ValueError('boolean index does not make sense')
        return Boolean
    elif not index and series_is_boolean(col):
        return Boolean
    elif is_integer_dtype(col):
        # parse purported 'integer' columns in a new table.
        # if values are all zero, make it a float for added safety - common case of a float that is often zero
        # if database table is type INTEGER, this will be parsed back to int later anyway
        if index:
            return Integer
        for val in col.unique():
            if pd.isna(val):
                continue
            if val != 0:
                return Integer
        return Float
    elif is_float_dtype(col):
        return Float
    elif is_datetime64_any_dtype(col):
        return TIMESTAMP(timezone=True)
    else:
        return String


def _get_type_from_db_col(col: sqa.Column):
    """Return Sqlalchemy type of a SQLAlchemy column. Updated to limit sqlalchemy types to this explicit list"""
    if isinstance(col.type, sqa.types.Integer):
        return Integer
    elif isinstance(col.type, sqa.types.Float):
        return Float
    elif isinstance(col.type, sqa.types.DateTime):
        return TIMESTAMP(timezone=True)
    elif isinstance(col.type, sqa.types.TIMESTAMP):
        return TIMESTAMP(timezone=True)
    elif isinstance(col.type, sqa.types.Boolean):
        return Boolean
    else:
        return String


def get_column_dtype(column, pd_or_sqla, index=False):
    """
    Take a column (sqlalchemy table.Column or df.Series), return its dtype in Pandas or SQLA

    If it doesn't match anything else, return String

    Args:
        column: pd.Series or SQLA.table.column
        pd_or_sqla: either 'pd' or 'sqla': which kind of type to return
        index: if True, column type cannot be boolean
    Returns:
        Type or None
            if pd_or_sqla == 'sqla':
                one of {Integer, Float, Boolean, DateTime, String, or None (for all-NaN column)}
            if pd_or_sqla == 'pd':
                one of {np.int64, np.float64, np.datetime64, np.bool_, np.str_}
    """
    if isinstance(column, sqa.Column):
        dtype = _get_type_from_db_col(column)
    elif isinstance(column, (pd.Series, pd.Index)):
        dtype = _get_type_from_df_col(column, index=index)
    else:
        raise ValueError(f'get_column_datatype takes a column; got {type(column)}')

    if dtype is None:
        return None

    elif pd_or_sqla == 'sqla':
        return dtype
    elif pd_or_sqla == 'pd':
        return _sqa_type2pandas_type(dtype, index=index)
    else:
        raise ValueError(f'Select pd_or_sqla must equal either "pd" or "sqla"')


def has_table(con, table_name, schema=None):
    """returns True if a table exactly named table_name exists in con"""
    engine = engine_builder(con)

    if schema is not None:
        return engine.run_callable(engine.dialect.has_table, table_name, schema=schema)
    else:
        return engine.run_callable(engine.dialect.has_table, table_name)


def clean_name(name):
    """returns a standardized version of name: lower case without spaces or special characters. May not be numeric"""
    d = {char: '_' for char in ' ()+-/*";=&|#><^%{}'}
    d['.'] = None
    d[','] = None
    table = str.maketrans(d)
    if name[0].isnumeric():
        raise NameError(f'Pandabase does not allow purely numeric names or names that start with digits.'
                        f' Illegal name: {name}')
    if '@' in name:
        raise NameError(f'At sign @ may not be a legal identifier. Please rename: {name}')
    return str(name).lower().strip().translate(table)


def make_clean_columns_dict(df: pd.DataFrame, autoindex=False):
    """Takes a DataFrame, returns a dictionary {column_names: {column info dicts}}

    if autoindex is True:
        if multi-index:
            fail; must reset_index first
        include an additional new Integer column named PANDABASE_DEFAULT_INDEX and discard df.index
    else:
        df.index.name (or .names for MultiIndex) must all be legal names (i.e. results of clean_name)

    Example:
        >>> import pandas as pd
        >>> data = {'full_name':['John Doe'],
        ...         'number_of_pets':[3],
        ...         'likes_bananas':[True], 
        ...         'dob':[pd.Timestamp('1990-01-01')]}
        >>> 
        >>> df = pd.DataFrame(data).rename_axis('id', axis = 'index')
        >>> make_clean_columns_dict(df)
        {'id': {'dtype': sqlalchemy.sql.sqltypes.Integer, 'pk': True},
         'full_name': {'dtype': sqlalchemy.sql.sqltypes.String, 'pk': False},
         'number_of_pets': {'dtype': sqlalchemy.sql.sqltypes.Integer, 'pk': False},
         'likes_bananas': {'dtype': sqlalchemy.sql.sqltypes.Boolean, 'pk': False},
         'dob': {'dtype': TIMESTAMP(timezone=True), 'pk': False}}
    
    """
    if len(df.columns) > 253:
        raise ValueError('pandabase is (currently) incompatible with data over 253 columns wide')
    columns = {}
    df.columns = [clean_name(col) for col in df.columns]

    # get index info
    if autoindex:
        if isinstance(df.index, pd.MultiIndex):
            raise ValueError(f'Must reset_index to use autoindex=True')

        index_name = PANDABASE_DEFAULT_INDEX
        columns[index_name] = {'dtype': Integer,
                               'pk': True}

    elif isinstance(df.index, pd.MultiIndex):
        indices = df[[]].reset_index(drop=False)
        for col_name in indices.columns:
            if col_name in df.columns:
                raise NameError(f'MultiIndex name is duplicate of column name: {col_name}')
            columns[col_name] = {'dtype': get_column_dtype(indices[col_name], 'sqla', index=True),
                                 'pk': True}
    else:
        if df.index.name in df.columns:
            raise NameError(f'index name is duplicate of column name: {df.index.name}')
        columns[df.index.name] = {'dtype': get_column_dtype(df.index, 'sqla', index=True),
                                  'pk': True}

    # get column info
    for col_name in df.columns:
        dtype = get_column_dtype(df[col_name], 'sqla')
        columns[col_name] = {
            'dtype': dtype,
            'pk': False,
        }
    assert len(columns) > 0

    return columns


def make_column(name, info):
    """Make a sqla.Column from column information; nullable unless it's the primary key"""
    nullable = not info['pk']
    return Column(name, primary_key=info['pk'], type_=info['dtype'], nullable=nullable)



from setuptools import setup

with open("README.md", "r") as readme:
    long_description = readme.read()

setup(
    version='0.4.5',
    author='Sam Beck',
    author_email='notsambeck@gmail.com',
    name='pandabase',
    packages=['pandabase'],
    description="pandabase links pandas DataFrames to SQL databases. Upsert, append, read, drop, describe...",
    long_description=long_description,
    long_description_content_type = 'text/markdown',
    url="https://github.com/notsambeck/pandabase",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        ],
    install_requires=[
        'pandas>=0.24.0',
        'sqlalchemy>=1.3.0',
        ],
    python_requires='>=3.6',
)



"""
pandabase converts pandas DataFrames to & from SQL databases

It replaces pandas.to_sql and pandas.read_sql, and requires the user
to select a unique index. This allows upserts and makes it easier to
maintain a dataset that grows over time. Especially time series.

pandabase:
    is simpler than pandas.io.sql
    is only compatible with newest versions of Pandas & sqlalchemy
    is not guaranteed
    definitely supports sqlite, may or may support other backends
    uses the sqlalchemy core and Pandas; has no additional dependencies

by sam beck
github.com/notsambeck/pandabase

largely copied from pandas:
https://github.com/pandas-dev/pandas
and dataset:
https://github.com/pudo/dataset/
"""
from pandabase.helpers import *

import pandas as pd
import numpy as np
from pandas.api.types import is_string_dtype

import sqlalchemy as sqa
from sqlalchemy import Table, and_
from sqlalchemy.dialects.postgresql import insert as pg_insert

import pytz
from logging import getLogger


logger = getLogger()


def to_sql(df: pd.DataFrame, *,
           table_name: str,
           con: str or sqa.engine,
           auto_index=False,
           how='create_only',
           add_new_columns=False, 
           schema: str = None):
    """
    Write records stored in DataFrame [df] to [table_name] in SQL database [con].

    Caveats:
    Converts any datetime to UTC
    Requires a unique, named index as DataFrame.index; to insert into existing database, this name must be consistent

    Args:
        df : DataFrame
        ### keyword only args follow ###
        table_name : string; Name of SQL table.
        con : connection; database string URI < OR > sqlalchemy.engine
        auto_index: bool, default False. if True, ignore existing df.index, make a new integer index
        add_new_columns: bool, default False. if True, add any new columns as required by the DataFrame.
        how : {'create_only', 'upsert', 'append'}, default 'create_only'
            - create_only:
                If table exists, raise an error and stop.
            - append:
                If table exists, append data. Raise if index overlaps
                Create table if does not exist.
            - upsert:
                create table if needed
                if record exists: update (possibly replacing values with NULL)
                else: insert
        schema: Specify the schema (if database flavor supports this, i.e. postgresql). If None, use default schema.
    
    """
    # 1. make connection objects
    df = df.copy()

    engine = engine_builder(con)
    meta = sqa.MetaData()

    ######################
    #
    #     2. validate inputs
    #
    ######################
    clean_table_name = clean_name(table_name)
    if clean_table_name != table_name:
        raise NameError(f'Illegal characters in table name: {table_name}. try: {clean_table_name}')

    if how not in ('create_only', 'append', 'upsert',):
        raise ValueError(f"Parameter how={how}; how must be in ['create_only', 'upsert', 'append']")

    if not isinstance(df, pd.DataFrame):
        raise ValueError('pandabase.to_sql() requires a DataFrame as input')

    if not auto_index:
        if not df.index.is_unique:
            raise ValueError('DataFrame.index is not unique and cannot be used as PK.')
        if is_datetime64_any_dtype(df.index):
            if df.index.tz != pytz.utc:
                raise ValueError(f'Index {df.index.name} is not UTC. Please correct.')

        if isinstance(df.index, pd.MultiIndex):
            alterations = {}
            for i, name in enumerate(df.index.names):
                if name is None:
                    raise NameError(f'One or more values in MultiIndex is unnamed: {df.index.names}')
                if clean_name(name) != name:
                    alterations[i] = clean_name(name)
            if alterations:
                new_names = list(df.index.names)
                for k, v in alterations.items():
                    new_names[k] = v
                df.index.names = new_names

        else:
            if df.index.name is None:
                raise NameError('Autoindex is False, but df.index.name is None. Set df.index.name = something')
            if df.index.hasnans:
                raise ValueError('DataFrame.index has NaN values and cannot be used as PK.')
            df.index.name = clean_name(df.index.name)

    else:
        if isinstance(df.index, pd.MultiIndex):
            raise ValueError(f'pandabase does not allow autoindex=True on a DataFrame with MultiIndex. '
                             f'Consider using df.reset_index(drop=[True or False]).')
        # otherwise: auto_index; drop index info
        df.reset_index(drop=True)
        df.index.name = PANDABASE_DEFAULT_INDEX

    for col in df.columns:
        if is_datetime64_any_dtype(df[col]):
            if df[col].dt.tz is None:
                raise ValueError(f'Column {col} timezone must be set, maybe with '
                                 f'col.tz_localize(pytz.timezone).tz_convert(pytz.utc)')
            elif df[col].dt.tz != pytz.utc:
                raise ValueError(f'Column {col} is set, but not to UTC. Maybe correct with col.tz_convert(pytz.utc)')

    # make a list of df columns for later:
    df_cols_dict = make_clean_columns_dict(df, autoindex=auto_index)

    #####################################################
    #
    #    3a. Make new Table from df info, if it does not exist...
    #
    #####################################################
    if not has_table(engine, table_name, schema=schema):
        
        # log the creation of the table
        if schema is not None:
            log_info = f'Creating new table {schema}.{table_name}'
        else:
            log_info = f'Creating new table {table_name}'
        
        logger.info(log_info)

        # create the table
        table = Table(table_name, 
                      meta,
                      *[make_column(name, info) for name, info in df_cols_dict.items()
                        if info['dtype'] is not None],
                      schema=schema)    # schema defaults to None also in the Table class

    #######################################################################################
    #
    #    3b. Or make Table from db schema
    #    db will be the source of truth for datatypes etc. in the future
    #
    #######################################################################################
    else:
        if how == 'create_only':
            raise NameError(f'Table {table_name} already exists; param "how" is set to "create_only".')

        table = Table(table_name, meta, autoload=True, autoload_with=engine, schema=schema)

        if how == 'upsert':
            if table.primary_key == PANDABASE_DEFAULT_INDEX or auto_index:
                raise IOError('Cannot upsert with an automatically generated index')

        # 3. iterate over df_columns; confirm that types are compatible and all columns exist
        for col_name, df_col_info in df_cols_dict.items():
            if col_name not in table.columns:
                if df_col_info['dtype'] is None:
                    continue   # skip empty columns that do not exist in db
                elif add_new_columns:
                    logger.info(f'adding new column to {con}:{table_name}: {col_name}')
                    _add_columns_to_db(make_column(col_name, df_col_info),
                                       table_name=table_name, con=con, schema=schema)
                    meta.clear()
                    table = Table(table_name, 
                                  meta, 
                                  autoload=True, 
                                  autoload_with=engine,
                                  schema=schema)  # schema defaults to None also in the Table class

                else:
                    raise NameError(f'New data has at least one column that does not exist in DB: {col_name}. \n'
                                    f'Set add_new_columns to True to automatically fix.')

            # check that dtypes and PKs match for existing columns
            col = table.columns[col_name]
            if col.primary_key != df_col_info['pk']:
                raise NameError(f'Inconsistent pk for col: {col_name}! db: {col.primary_key} / '
                                f'df: {df_col_info["pk"]}')

            db_sqla_dtype = get_column_dtype(col, pd_or_sqla='sqla')

            # 3c. check datatypes
            if db_sqla_dtype == df_col_info['dtype']:
                continue

            #############################################
            #
            #    3d. COERCE INCONSISTENT DATATYPES - case by case
            #    try to change DataFrame dtypes to match existing db dtypes
            #
            #############################################
            db_pandas_dtype = get_column_dtype(col, pd_or_sqla='pd')

            if df_col_info['dtype'] is None:  # dtype is None only if the whole column is NULL
                # this does not need to be explicitly handled because when inserting None, nothing happens
                continue

            elif is_datetime64_any_dtype(db_pandas_dtype):
                pass
                # TODO: df[col_name] = pd.to_datetime(df[col_name].values, utc=True)
                # TODO: would be nice, but THIS FAILS FOR INDEX/PK. would solve cases when:
                # db.col is datetime and df.col is string
                # df.datetime_column.tz != utc

            elif (
                    df_col_info['dtype'] == Integer and is_float_dtype(db_pandas_dtype) or
                    df_col_info['dtype'] == Float and is_integer_dtype(db_pandas_dtype)
            ):
                # print(f'NUMERIC DTYPE: converting df[{name}] from {df[name].dtype} to {db_pandas_dtype}')
                try:
                    df[col_name] = df[col_name].astype(db_pandas_dtype)
                except Exception as e:
                    raise TypeError(f'Error {e} while trying to coerce float to int or vice versa, '
                                    f'e.g. cannot coerce {df[col_name][0]} to {db_pandas_dtype}')
            elif (
                    df_col_info['dtype'] == Boolean and is_integer_dtype(db_pandas_dtype) or
                    df_col_info['dtype'] == Boolean and is_float_dtype(db_pandas_dtype)
            ):
                try:
                    df[col_name] = df[col_name].astype(db_pandas_dtype)
                except Exception as e:
                    raise TypeError(f'Error {e} while trying to coerce bool float or int, '
                                    f'e.g. cannot coerce {df[col_name][0]} to {db_pandas_dtype}')

            else:
                raise TypeError(
                    f'Inconsistent type for column: {col_name} \n'
                    f'db.{col_name}.dtype= {db_pandas_dtype} / '
                    f'df{col_name}.dtype= {df_col_info["dtype"]}')

    #######################################################
    # FINALLY: either insert/fail, append/fail, or upsert #
    #######################################################

    # print('DB connection begins...')
    with engine.begin() as con:
        meta.create_all(bind=con)

    if how in ['append', 'create_only']:
        # will raise IntegrityError if repeated index encountered
        _insert(table, engine, df, auto_index)

    elif how == 'upsert':
        _upsert(table, engine, df)

    return table


def _insert(table: sqa.Table,
            engine: sqa.engine,
            cleaned_data: pd.DataFrame,
            auto_index: bool):

    with engine.begin() as con:
        rows = []
        
        # remove completely null columns; convert to object due to bug inserting Int64
        df = cleaned_data.dropna(axis=1, how='all').astype('object')

        if not auto_index:
            for row in df.reset_index(drop=False).itertuples(index=False):
                rows.append(row._asdict())
            con.execute(table.insert(), rows)
        else:
            for row in df.reset_index(drop=True).itertuples(index=False):
                rows.append(row._asdict())
            con.execute(table.insert(), rows)


def _upsert(table: sqa.Table,
            engine: sqa.engine,
            cleaned_data: pd.DataFrame):
    """
    insert data into a table, replacing any rows with duplicate indices

    When upsert finds duplicate records, it overwrites ALL VALUES that are present in source DataFrame, including NaN.

    postgres - see: https://docs.sqlalchemy.org/en/13/dialects/postgresql.html#insert-on-conflict-upsert
    """
    if isinstance(cleaned_data.index, pd.MultiIndex):
        multi = True
        names = cleaned_data.index.names
        index_elements = names
    else:
        multi = False
        names = cleaned_data.index.name
        index_elements = [names]

    cleaned_data = cleaned_data.astype('object')

    def map2none(val):
        """sqlalchemy freaks out about NaNs, so replace them with None"""
        if pd.notna(val):
            return val

    with engine.begin() as con:
        for row in cleaned_data.reset_index(drop=False).itertuples(index=False):
            # check index uniqueness by attempting insert; if it fails, update
            row = {k: map2none(v) for k, v in row._asdict().items()}
            try:

                if engine.dialect.dbapi.__name__ == 'psycopg2':
                    insert = pg_insert(table).values(row).on_conflict_do_update(
                        index_elements=index_elements,
                        set_=row
                    )
                else:
                    insert = table.insert().values(row)

                con.execute(insert)

            except sqa.exc.IntegrityError:
                if multi:
                    upsert = table.update()
                    for n in names:
                        upsert = upsert.where(table.c[n] == row[n])
                    con.execute(upsert.values(row))
                else:
                    upsert = table.update() \
                        .where(table.c[names] == row[names]) \
                        .values(row)
                    con.execute(upsert)


def read_sql(table_name: str,
             con: str or sqa.engine,
             *,
             lowest=None, highest=None, schema: str = None):
    """
    Read in a table from con as a pd.DataFrame, preserving dtypes and primary keys

    Args:
        table_name:
        con:
        lowest: minimum value of PK to select (inclusive)
        highest: maximum value of PK to select (inclusive)
            for MultiIndex tables, highest and lowest must be tuples of all primary key values (left to right) or None
        schema: Specify the schema (if database flavor supports this). If None, use default schema.

    Returns:
        DataFrame: selected data with table.primary_key(s) as index
    """
    engine = engine_builder(con)
    meta = sqa.MetaData(bind=engine)
    table = Table(table_name, 
                  meta, 
                  autoload=True, 
                  autoload_with=engine, 
                  schema=schema)   # schema defaults to None also in the Table class

    if len(table.primary_key.columns) == 0:
        print(f'Table {table_name} has no explicit PK/index (using autoindex)')
        assert lowest is None
        assert highest is None
        result = engine.execute(table.select())
        data = result.fetchall()

    elif len(table.primary_key.columns) == 1:
        pk = table.primary_key.columns.items()[0][1]

        if highest is None:
            if lowest is None:
                s = table.select()
            else:
                s = table.select().where(pk >= lowest)
        else:
            if lowest is None:
                s = table.select().where(pk <= highest)
            else:
                s = table.select().where(and_(pk >= lowest,
                                              pk <= highest))
        result = engine.execute(s)
        data = result.fetchall()

        if len(data) == 0:
            if not isinstance(lowest, pk.type.python_type) or not isinstance(highest, pk.type.python_type):
                raise TypeError(f'Select range is: {lowest} <= data <= {highest}; but type of column is {pk.type}. '
                                f'If e.g. cutoffs are integers, and column is float, please coerce type.')

    elif len(table.primary_key.columns) > 1:
        pks = table.primary_key.columns.items()

        s = table.select()

        for selector, sign in [(highest, 'highest'), (lowest, 'lowest')]:
            if selector is None:
                continue

            if len(selector) != len(pks):
                raise ValueError('pandabase.read_sql(multi-indexed_table) requires any values of highest, lowest '
                                 'that are not None to have __len__ equal to len(index). Use e.g. (value, None) to '
                                 'filter only on the first index dimension.')

            for i, val in enumerate(selector):
                if val is None:
                    continue
                elif sign == 'lowest':
                    s = s.where(pks[i][1] >= val)
                else:
                    s = s.where(pks[i][1] <= val)

        result = engine.execute(s)
        data = result.fetchall()

        if len(data) == 0:
            for i, (pk, col) in enumerate(table.primary_key.columns.items()):
                if isinstance(lowest[i], col.type.python_type) or isinstance(highest[i], col.type.python_type):
                    continue
                else:
                    raise TypeError(f'Select range is: {lowest[i]} <= data <= {highest[i]}; '
                                    f'but type of column is {col.type.python_type}')

    else:
        raise ValueError(f'invalid table.primary_key.columns = {table.primary_key.columns}')

    df = pd.DataFrame.from_records(data, columns=[col.name for col in table.columns],
                                   coerce_float=True)

    indices = []  # in case of multi-index, accumulate columns and assemble later
    for col in table.columns:
        # deal with primary keys separately; never convert primary key to nullable
        if col.primary_key:
            dtype = get_column_dtype(col, pd_or_sqla='pd', index=True)

            # single index
            if len(table.primary_key.columns) == 1:
                df.index = df[col.name]

                if is_datetime64_any_dtype(dtype):
                    df.index = pd.to_datetime(df[col.name].values, utc=True)

                # auto-handle legacy index name with random number suffix
                if col.name == PANDABASE_DEFAULT_INDEX or col.name[:23] == 'pandabase_default_index':
                    df.index.name = None
                else:
                    df.index.name = col.name

                df = df.drop(columns=[col.name])
                continue

            # multi-index
            else:
                indices.append(df[col.name].copy())

                if is_datetime64_any_dtype(dtype):
                    indices[-1] = pd.to_datetime(indices[-1].values, utc=True)

                df = df.drop(columns=[col.name])
                continue
        else:
            # print(f'non-pk column: {col}')
            dtype = get_column_dtype(col, pd_or_sqla='pd')
            # force all dates to utc
            if is_datetime64_any_dtype(dtype):
                # print(df[col.name].dt.tz, 'regular col - old...')
                df[col.name] = pd.to_datetime(df[col.name].values, utc=True)
                # print(df[col.name].dt.tz, 'regular col - new')

        # convert other dtypes to nullable
        if is_bool_dtype(dtype) or is_integer_dtype(dtype):
            df[col.name] = np.array(df[col.name], dtype=float)
            df[col.name] = df[col.name].astype(pd.Int64Dtype())
        elif is_float_dtype(dtype):
            pass
        elif is_string_dtype(col):
            pass

    # generate multi_index
    if indices:
        df.index = pd.MultiIndex.from_arrays(indices)

    return df


def _add_columns_to_db(new_col, table_name, con, schema=None):
    """Make new columns as needed with ALTER TABLE, as a weak substitute for migrations"""
    engine = engine_builder(con)
    name = clean_name(new_col.name)
    
    if schema is None:
        table_namespace = table_name
    else:
        table_namespace = f'{schema}.{table_name}'

    with engine.begin() as conn:
        conn.execute(f'ALTER TABLE {table_namespace} '
                     f'ADD COLUMN {name} {new_col.type.compile(engine.dialect)}')


def profiling_script(n_rows):
    db = 'sqlite:///:memory:'
    n_cols = 100
    _df = pd.DataFrame(index=range(n_rows), columns=['col_' + str(n) for n in range(n_cols)],
                       data=np.random.random((n_rows, n_cols)))
    _df.index.name = 'dex'
    to_sql(_df, con=db, table_name='table', how='append')


if __name__ == '__main__':
    profiling_script(10000)
